# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/053_optimizer.ipynb.

# %% auto 0
__all__ = ['wrap_optimizer']

# %% ../nbs/053_optimizer.ipynb 4
from functools import partial
from fastai.optimizer import *

# %% ../nbs/053_optimizer.ipynb 5
def wrap_optimizer(opt, **kwargs): return partial(OptimWrapper, opt=opt, **kwargs)
