# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_nb_name": "000_export.ipynb",
         "get_colab_nb_name": "000_export.ipynb",
         "get_nb_path": "000_export.ipynb",
         "nb_name_to_py": "000_export.ipynb",
         "get_script_path": "000_export.ipynb",
         "nb2py": "000_export.ipynb",
         "is_nparray": "001_utils.ipynb",
         "is_tensor": "001_utils.ipynb",
         "is_zarr": "001_utils.ipynb",
         "is_dask": "001_utils.ipynb",
         "is_memmap": "001_utils.ipynb",
         "is_slice": "001_utils.ipynb",
         "totensor": "001_utils.ipynb",
         "toarray": "001_utils.ipynb",
         "toL": "001_utils.ipynb",
         "to3dtensor": "001_utils.ipynb",
         "to2dtensor": "001_utils.ipynb",
         "to1dtensor": "001_utils.ipynb",
         "to3darray": "001_utils.ipynb",
         "to2darray": "001_utils.ipynb",
         "to1darray": "001_utils.ipynb",
         "to3d": "001_utils.ipynb",
         "to2d": "001_utils.ipynb",
         "to1d": "001_utils.ipynb",
         "to2dPlus": "001_utils.ipynb",
         "to3dPlus": "001_utils.ipynb",
         "to2dPlusTensor": "001_utils.ipynb",
         "to2dPlusArray": "001_utils.ipynb",
         "to3dPlusTensor": "001_utils.ipynb",
         "to3dPlusArray": "001_utils.ipynb",
         "todtype": "001_utils.ipynb",
         "bytes2size": "001_utils.ipynb",
         "bytes2GB": "001_utils.ipynb",
         "get_size": "001_utils.ipynb",
         "get_dir_size": "001_utils.ipynb",
         "get_file_size": "001_utils.ipynb",
         "is_file": "001_utils.ipynb",
         "is_dir": "001_utils.ipynb",
         "delete_all_in_dir": "001_utils.ipynb",
         "reverse_dict": "001_utils.ipynb",
         "is_tuple": "001_utils.ipynb",
         "itemify": "001_utils.ipynb",
         "isnone": "001_utils.ipynb",
         "exists": "001_utils.ipynb",
         "ifelse": "001_utils.ipynb",
         "is_not_close": "001_utils.ipynb",
         "test_not_close": "001_utils.ipynb",
         "test_type": "001_utils.ipynb",
         "test_ok": "001_utils.ipynb",
         "test_not_ok": "001_utils.ipynb",
         "test_error": "001_utils.ipynb",
         "test_eq_nan": "001_utils.ipynb",
         "assert_fn": "001_utils.ipynb",
         "test_gt": "001_utils.ipynb",
         "test_ge": "001_utils.ipynb",
         "test_lt": "001_utils.ipynb",
         "test_le": "001_utils.ipynb",
         "stack": "001_utils.ipynb",
         "stack_pad": "001_utils.ipynb",
         "pad_sequences": "001_utils.ipynb",
         "match_seq_len": "001_utils.ipynb",
         "random_shuffle": "001_utils.ipynb",
         "cat2int": "001_utils.ipynb",
         "cycle_dl": "001_utils.ipynb",
         "cycle_dl_to_device": "001_utils.ipynb",
         "cycle_dl_estimate": "001_utils.ipynb",
         "cache_data": "001_utils.ipynb",
         "memmap2cache": "001_utils.ipynb",
         "cache_memmap": "001_utils.ipynb",
         "get_func_defaults": "001_utils.ipynb",
         "get_idx_from_df_col_vals": "001_utils.ipynb",
         "get_sublist_idxs": "001_utils.ipynb",
         "flatten_list": "001_utils.ipynb",
         "display_pd_df": "001_utils.ipynb",
         "ttest": "001_utils.ipynb",
         "kstest": "001_utils.ipynb",
         "tscore": "001_utils.ipynb",
         "pcc": "001_utils.ipynb",
         "scc": "001_utils.ipynb",
         "a": "001_utils.ipynb",
         "b": "001_utils.ipynb",
         "remove_fn": "001_utils.ipynb",
         "npsave": "001_utils.ipynb",
         "np_save": "001_utils.ipynb",
         "permute_2D": "001_utils.ipynb",
         "random_normal": "001_utils.ipynb",
         "random_half_normal": "001_utils.ipynb",
         "random_normal_tensor": "001_utils.ipynb",
         "random_half_normal_tensor": "001_utils.ipynb",
         "default_dpi": "001_utils.ipynb",
         "get_plot_fig": "001_utils.ipynb",
         "fig2buf": "001_utils.ipynb",
         "plot_scatter": "001_utils.ipynb",
         "get_idxs": "001_utils.ipynb",
         "apply_cmap": "001_utils.ipynb",
         "torch_tile": "001_utils.ipynb",
         "to_tsfresh_df": "001_utils.ipynb",
         "pcorr": "001_utils.ipynb",
         "scorr": "001_utils.ipynb",
         "torch_diff": "001_utils.ipynb",
         "get_outliers_IQR": "001_utils.ipynb",
         "clip_outliers": "001_utils.ipynb",
         "get_percentile": "001_utils.ipynb",
         "torch_clamp": "001_utils.ipynb",
         "get_robustscale_params": "001_utils.ipynb",
         "torch_slice_by_dim": "001_utils.ipynb",
         "torch_nanmean": "001_utils.ipynb",
         "torch_nanstd": "001_utils.ipynb",
         "concat": "001_utils.ipynb",
         "reduce_memory_usage": "001_utils.ipynb",
         "cls_name": "001_utils.ipynb",
         "roll2d": "001_utils.ipynb",
         "roll3d": "001_utils.ipynb",
         "random_roll2d": "001_utils.ipynb",
         "random_roll3d": "001_utils.ipynb",
         "rotate_axis0": "001_utils.ipynb",
         "rotate_axis1": "001_utils.ipynb",
         "rotate_axis2": "001_utils.ipynb",
         "chunks_calculator": "001_utils.ipynb",
         "is_memory_shared": "001_utils.ipynb",
         "assign_in_chunks": "001_utils.ipynb",
         "create_array": "001_utils.ipynb",
         "create_empty_array": "001_utils.ipynb",
         "np_save_compressed": "001_utils.ipynb",
         "np_load_compressed": "001_utils.ipynb",
         "np2memmap": "001_utils.ipynb",
         "torch_mean_groupby": "001_utils.ipynb",
         "torch_flip": "001_utils.ipynb",
         "torch_nan_to_num": "001_utils.ipynb",
         "torch_masked_to_num": "001_utils.ipynb",
         "mpl_trend": "001_utils.ipynb",
         "int2digits": "001_utils.ipynb",
         "array2digits": "001_utils.ipynb",
         "sincos_encoding": "001_utils.ipynb",
         "linear_encoding": "001_utils.ipynb",
         "encode_positions": "001_utils.ipynb",
         "sort_generator": "001_utils.ipynb",
         "get_subset_dict": "001_utils.ipynb",
         "create_dir": "001_utils.ipynb",
         "remove_dir": "001_utils.ipynb",
         "named_partial": "001_utils.ipynb",
         "yaml2dict": "001_utils.ipynb",
         "str2list": "001_utils.ipynb",
         "str2index": "001_utils.ipynb",
         "get_cont_cols": "001_utils.ipynb",
         "get_cat_cols": "001_utils.ipynb",
         "alphabet": "001_utils.ipynb",
         "ALPHABET": "001_utils.ipynb",
         "get_mapping": "001_utils.ipynb",
         "map_array": "001_utils.ipynb",
         "log_tfm": "001_utils.ipynb",
         "to_sincos_time": "001_utils.ipynb",
         "plot_feature_dist": "001_utils.ipynb",
         "rolling_moving_average": "001_utils.ipynb",
         "ffill_sequence": "001_utils.ipynb",
         "bfill_sequence": "001_utils.ipynb",
         "fbfill_sequence": "001_utils.ipynb",
         "dummify": "001_utils.ipynb",
         "shuffle_along_axis": "001_utils.ipynb",
         "analyze_feature": "001_utils.ipynb",
         "analyze_array": "001_utils.ipynb",
         "get_relpath": "001_utils.ipynb",
         "split_in_chunks": "001_utils.ipynb",
         "save_object": "001_utils.ipynb",
         "load_object": "001_utils.ipynb",
         "get_idxs_to_keep": "001_utils.ipynb",
         "zerofy": "001_utils.ipynb",
         "RandomSplitter": "010_data.validation.ipynb",
         "check_overlap": "010_data.validation.ipynb",
         "check_splits_overlap": "010_data.validation.ipynb",
         "leakage_finder": "010_data.validation.ipynb",
         "balance_idx": "010_data.validation.ipynb",
         "TrainValidTestSplitter": "010_data.validation.ipynb",
         "plot_splits": "010_data.validation.ipynb",
         "get_splits": "010_data.validation.ipynb",
         "TSSplitter": "010_data.validation.ipynb",
         "TimeSplitter": "010_data.validation.ipynb",
         "get_predefined_splits": "010_data.validation.ipynb",
         "combine_split_data": "010_data.validation.ipynb",
         "get_splits_len": "010_data.validation.ipynb",
         "prepare_idxs": "011_data.preparation.ipynb",
         "prepare_sel_vars_and_steps": "011_data.preparation.ipynb",
         "apply_sliding_window": "011_data.preparation.ipynb",
         "df2Xy": "011_data.preparation.ipynb",
         "split_Xy": "011_data.preparation.ipynb",
         "df2xy": "011_data.preparation.ipynb",
         "split_xy": "011_data.preparation.ipynb",
         "df2np3d": "011_data.preparation.ipynb",
         "add_missing_value_cols": "011_data.preparation.ipynb",
         "add_missing_timestamps": "011_data.preparation.ipynb",
         "time_encoding": "011_data.preparation.ipynb",
         "forward_gaps": "011_data.preparation.ipynb",
         "backward_gaps": "011_data.preparation.ipynb",
         "nearest_gaps": "011_data.preparation.ipynb",
         "get_gaps": "011_data.preparation.ipynb",
         "add_delta_timestamp_cols": "011_data.preparation.ipynb",
         "SlidingWindow": "011_data.preparation.ipynb",
         "SlidingWindowSplitter": "011_data.preparation.ipynb",
         "SlidingWindowPanel": "011_data.preparation.ipynb",
         "SlidingWindowPanelSplitter": "011_data.preparation.ipynb",
         "identify_padding": "011_data.preparation.ipynb",
         "decompress_from_url": "012_data.external.ipynb",
         "download_data": "012_data.external.ipynb",
         "get_UCR_univariate_list": "012_data.external.ipynb",
         "UTSC_datasets": "012_data.external.ipynb",
         "UCR_univariate_list": "012_data.external.ipynb",
         "get_UCR_multivariate_list": "012_data.external.ipynb",
         "MTSC_datasets": "012_data.external.ipynb",
         "UCR_multivariate_list": "012_data.external.ipynb",
         "UCR_list": "012_data.external.ipynb",
         "classification_list": "012_data.external.ipynb",
         "TSC_datasets": "012_data.external.ipynb",
         "get_UCR_data": "012_data.external.ipynb",
         "get_classification_data": "012_data.external.ipynb",
         "check_data": "012_data.external.ipynb",
         "get_Monash_regression_list": "012_data.external.ipynb",
         "Monash_regression_list": "012_data.external.ipynb",
         "regression_list": "012_data.external.ipynb",
         "TSR_datasets": "012_data.external.ipynb",
         "get_Monash_regression_data": "012_data.external.ipynb",
         "get_regression_data": "012_data.external.ipynb",
         "get_forecasting_list": "012_data.external.ipynb",
         "forecasting_time_series": "012_data.external.ipynb",
         "get_forecasting_time_series": "012_data.external.ipynb",
         "Monash_forecasting_list": "012_data.external.ipynb",
         "forecasting_list": "012_data.external.ipynb",
         "convert_tsf_to_dataframe": "012_data.external.ipynb",
         "get_Monash_forecasting_data": "012_data.external.ipynb",
         "get_forecasting_data": "012_data.external.ipynb",
         "NumpyTensor": "013_data.core.ipynb",
         "ToNumpyTensor": "013_data.core.ipynb",
         "TSTensor": "013_data.core.ipynb",
         "ToTSTensor": "013_data.core.ipynb",
         "show_tuple": "013_data.core.ipynb",
         "TSLabelTensor": "013_data.core.ipynb",
         "TSMaskTensor": "013_data.core.ipynb",
         "ToFloat": "013_data.core.ipynb",
         "ToInt": "013_data.core.ipynb",
         "TSClassification": "013_data.core.ipynb",
         "TSCategorize": "013_data.core.ipynb",
         "TSRegression": "013_data.core.ipynb",
         "TSForecasting": "013_data.core.ipynb",
         "TSMultiLabelClassification": "013_data.core.ipynb",
         "NumpyTensorBlock": "013_data.core.ipynb",
         "TSTensorBlock": "013_data.core.ipynb",
         "TorchDataset": "013_data.core.ipynb",
         "NumpyDataset": "013_data.core.ipynb",
         "TSDataset": "013_data.core.ipynb",
         "NoTfmLists": "013_data.core.ipynb",
         "TSTfmdLists": "013_data.core.ipynb",
         "NumpyDatasets": "013_data.core.ipynb",
         "tscoll_repr": "013_data.core.ipynb",
         "TSDatasets": "013_data.core.ipynb",
         "add_ds": "013_data.core.ipynb",
         "NumpyDatasets.add_dataset": "013_data.core.ipynb",
         "NumpyDatasets.add_test": "013_data.core.ipynb",
         "NumpyDatasets.add_unlabeled": "013_data.core.ipynb",
         "NumpyDataLoader": "013_data.core.ipynb",
         "TSDataLoader": "013_data.core.ipynb",
         "NumpyDataLoaders": "013_data.core.ipynb",
         "TSDataLoaders": "013_data.core.ipynb",
         "StratifiedSampler": "013_data.core.ipynb",
         "get_c": "013_data.core.ipynb",
         "get_best_dl_params": "013_data.core.ipynb",
         "get_best_dls_params": "013_data.core.ipynb",
         "get_ts_dls": "013_data.core.ipynb",
         "get_ts_dl": "013_data.core.ipynb",
         "get_subset_dl": "013_data.core.ipynb",
         "get_tsimage_dls": "013_data.core.ipynb",
         "get_time_per_batch": "013_data.core.ipynb",
         "get_dl_percent_per_epoch": "013_data.core.ipynb",
         "TSUnwindowedDataset": "014_data.unwindowed.ipynb",
         "TSUnwindowedDatasets": "014_data.unwindowed.ipynb",
         "TSMetaDataset": "015_data.metadatasets.ipynb",
         "TSMetaDatasets": "015_data.metadatasets.ipynb",
         "ToNumpyCategory": "016_data.preprocessing.ipynb",
         "OneHot": "016_data.preprocessing.ipynb",
         "TSNan2Value": "016_data.preprocessing.ipynb",
         "Nan2Value": "016_data.preprocessing.ipynb",
         "TSStandardize": "016_data.preprocessing.ipynb",
         "torch.Tensor.mul_min": "016_data.preprocessing.ipynb",
         "TSTensor.mul_min": "016_data.preprocessing.ipynb",
         "NumpyTensor.mul_min": "016_data.preprocessing.ipynb",
         "torch.Tensor.mul_max": "016_data.preprocessing.ipynb",
         "TSTensor.mul_max": "016_data.preprocessing.ipynb",
         "NumpyTensor.mul_max": "016_data.preprocessing.ipynb",
         "TSNormalize": "016_data.preprocessing.ipynb",
         "TSClipOutliers": "016_data.preprocessing.ipynb",
         "TSClip": "016_data.preprocessing.ipynb",
         "TSSelfMissingness": "016_data.preprocessing.ipynb",
         "TSRobustScale": "016_data.preprocessing.ipynb",
         "get_stats_with_uncertainty": "016_data.preprocessing.ipynb",
         "get_random_stats": "016_data.preprocessing.ipynb",
         "TSGaussianStandardize": "016_data.preprocessing.ipynb",
         "TSRandomStandardize": "016_data.preprocessing.ipynb",
         "TSDiff": "016_data.preprocessing.ipynb",
         "TSLog": "016_data.preprocessing.ipynb",
         "TSCyclicalPosition": "016_data.preprocessing.ipynb",
         "TSLinearPosition": "016_data.preprocessing.ipynb",
         "TSMissingness": "016_data.preprocessing.ipynb",
         "TSPositionGaps": "016_data.preprocessing.ipynb",
         "TSRollingMean": "016_data.preprocessing.ipynb",
         "TSLogReturn": "016_data.preprocessing.ipynb",
         "TSAdd": "016_data.preprocessing.ipynb",
         "TSClipByVar": "016_data.preprocessing.ipynb",
         "TSDropVars": "016_data.preprocessing.ipynb",
         "TSOneHotEncode": "016_data.preprocessing.ipynb",
         "TSPosition": "016_data.preprocessing.ipynb",
         "TSShrinkDataFrame": "016_data.preprocessing.ipynb",
         "TSOneHotEncoder": "016_data.preprocessing.ipynb",
         "TSCategoricalEncoder": "016_data.preprocessing.ipynb",
         "TSDateTimeEncoder": "016_data.preprocessing.ipynb",
         "default_date_attr": "016_data.preprocessing.ipynb",
         "TSMissingnessEncoder": "016_data.preprocessing.ipynb",
         "Preprocessor": "016_data.preprocessing.ipynb",
         "StandardScaler": "016_data.preprocessing.ipynb",
         "RobustScaler": "016_data.preprocessing.ipynb",
         "Normalizer": "016_data.preprocessing.ipynb",
         "BoxCox": "016_data.preprocessing.ipynb",
         "YeoJohnshon": "016_data.preprocessing.ipynb",
         "Quantile": "016_data.preprocessing.ipynb",
         "ReLabeler": "016_data.preprocessing.ipynb",
         "TSIdentity": "017_data.transforms.ipynb",
         "TSShuffle_HLs": "017_data.transforms.ipynb",
         "TSShuffleSteps": "017_data.transforms.ipynb",
         "TSGaussianNoise": "017_data.transforms.ipynb",
         "TSMagAddNoise": "017_data.transforms.ipynb",
         "TSMagMulNoise": "017_data.transforms.ipynb",
         "random_curve_generator": "017_data.transforms.ipynb",
         "random_cum_curve_generator": "017_data.transforms.ipynb",
         "random_cum_noise_generator": "017_data.transforms.ipynb",
         "random_cum_linear_generator": "017_data.transforms.ipynb",
         "TSTimeNoise": "017_data.transforms.ipynb",
         "TSMagWarp": "017_data.transforms.ipynb",
         "TSTimeWarp": "017_data.transforms.ipynb",
         "TSWindowWarp": "017_data.transforms.ipynb",
         "TSMagScale": "017_data.transforms.ipynb",
         "TSMagScalePerVar": "017_data.transforms.ipynb",
         "TSMagScaleByVar": "017_data.transforms.ipynb",
         "TSRandomResizedCrop": "017_data.transforms.ipynb",
         "TSRandomZoomIn": "017_data.transforms.ipynb",
         "TSWindowSlicing": "017_data.transforms.ipynb",
         "TSRandomZoomOut": "017_data.transforms.ipynb",
         "TSRandomTimeScale": "017_data.transforms.ipynb",
         "TSRandomTimeStep": "017_data.transforms.ipynb",
         "TSResampleSteps": "017_data.transforms.ipynb",
         "TSSubsampleSteps": "017_data.transforms.ipynb",
         "TSBlur": "017_data.transforms.ipynb",
         "TSSmooth": "017_data.transforms.ipynb",
         "maddest": "017_data.transforms.ipynb",
         "TSFreqDenoise": "017_data.transforms.ipynb",
         "TSRandomFreqNoise": "017_data.transforms.ipynb",
         "TSRandomResizedLookBack": "017_data.transforms.ipynb",
         "TSRandomLookBackOut": "017_data.transforms.ipynb",
         "TSVarOut": "017_data.transforms.ipynb",
         "TSCutOut": "017_data.transforms.ipynb",
         "TSTimeStepOut": "017_data.transforms.ipynb",
         "TSRandomCropPad": "017_data.transforms.ipynb",
         "TSMaskOut": "017_data.transforms.ipynb",
         "TSInputDropout": "017_data.transforms.ipynb",
         "TSTranslateX": "017_data.transforms.ipynb",
         "TSRandomShift": "017_data.transforms.ipynb",
         "TSHorizontalFlip": "017_data.transforms.ipynb",
         "TSRandomTrend": "017_data.transforms.ipynb",
         "TSRandomRotate": "017_data.transforms.ipynb",
         "TSVerticalFlip": "017_data.transforms.ipynb",
         "TSResize": "017_data.transforms.ipynb",
         "TSRandomSize": "017_data.transforms.ipynb",
         "TSRandomLowRes": "017_data.transforms.ipynb",
         "TSDownUpScale": "017_data.transforms.ipynb",
         "TSRandomDownUpScale": "017_data.transforms.ipynb",
         "TSRandomConv": "017_data.transforms.ipynb",
         "TSRandom2Value": "017_data.transforms.ipynb",
         "TSMask2Value": "017_data.transforms.ipynb",
         "all_TS_randaugs": "017_data.transforms.ipynb",
         "RandAugment": "017_data.transforms.ipynb",
         "TestTfm": "017_data.transforms.ipynb",
         "get_tfm_name": "017_data.transforms.ipynb",
         "MixHandler1d": "018_data.mixed_augmentation.ipynb",
         "MixUp1d": "018_data.mixed_augmentation.ipynb",
         "MixUp1D": "018_data.mixed_augmentation.ipynb",
         "CutMix1d": "018_data.mixed_augmentation.ipynb",
         "IntraClassCutMix1d": "018_data.mixed_augmentation.ipynb",
         "TSImage": "019_data.image.ipynb",
         "ToTSImage": "019_data.image.ipynb",
         "TSToPlot": "019_data.image.ipynb",
         "TSToMat": "019_data.image.ipynb",
         "TSToGADF": "019_data.image.ipynb",
         "TSToGASF": "019_data.image.ipynb",
         "TSToMTF": "019_data.image.ipynb",
         "TSToRP": "019_data.image.ipynb",
         "TSToJRP": "019_data.image.ipynb",
         "get_ts_features": "020_data.features.ipynb",
         "get_tabular_ds": "021_data.tabular.ipynb",
         "get_tabular_dls": "021_data.tabular.ipynb",
         "preprocess_df": "021_data.tabular.ipynb",
         "MixedDataLoader": "022_data.mixed.ipynb",
         "MixedDataLoaders": "022_data.mixed.ipynb",
         "get_mixed_dls": "022_data.mixed.ipynb",
         "HuberLoss": "050_losses.ipynb",
         "LogCoshLoss": "050_losses.ipynb",
         "MaskedLossWrapper": "050_losses.ipynb",
         "CenterLoss": "050_losses.ipynb",
         "CenterPlusLoss": "050_losses.ipynb",
         "FocalLoss": "050_losses.ipynb",
         "TweedieLoss": "050_losses.ipynb",
         "MatthewsCorrCoefBinary": "051_metrics.ipynb",
         "get_task_metrics": "051_metrics.ipynb",
         "accuracy_multi": "051_metrics.ipynb",
         "metrics_multi_common": "051_metrics.ipynb",
         "precision_multi": "051_metrics.ipynb",
         "recall_multi": "051_metrics.ipynb",
         "specificity_multi": "051_metrics.ipynb",
         "balanced_accuracy_multi": "051_metrics.ipynb",
         "Fbeta_multi": "051_metrics.ipynb",
         "F1_multi": "051_metrics.ipynb",
         "mae": "051_metrics.ipynb",
         "mape": "051_metrics.ipynb",
         "recall_at_specificity": "051_metrics.ipynb",
         "lift": "051_metrics.ipynb",
         "lift_at_specificity": "051_metrics.ipynb",
         "top_k_lift": "051_metrics.ipynb",
         "mean_per_class_accuracy": "051_metrics.ipynb",
         "Learner.show_batch": "052_learner.ipynb",
         "Learner.remove_all_cbs": "052_learner.ipynb",
         "Learner.one_batch": "052_learner.ipynb",
         "Learner.save_all": "052_learner.ipynb",
         "load_all": "052_learner.ipynb",
         "load_learner_all": "052_learner.ipynb",
         "Recorder.plot_metrics": "052_learner.ipynb",
         "Learner.plot_metrics": "052_learner.ipynb",
         "get_arch": "052_learner.ipynb",
         "all_archs_names": "052_learner.ipynb",
         "ts_learner": "052_learner.ipynb",
         "tsimage_learner": "052_learner.ipynb",
         "Learner.decoder": "052_learner.ipynb",
         "Learner.get_X_preds": "052a_inference.ipynb",
         "Learner.show_probas": "052b_analysis.ipynb",
         "Learner.plot_confusion_matrix": "052b_analysis.ipynb",
         "Learner.top_losses": "052b_analysis.ipynb",
         "Learner.plot_top_losses": "052b_analysis.ipynb",
         "Learner.feature_importance": "052b_analysis.ipynb",
         "Learner.step_importance": "052b_analysis.ipynb",
         "ModelWithTemperature": "052c_calibration.ipynb",
         "TemperatureSetter": "052c_calibration.ipynb",
         "ECELoss": "052c_calibration.ipynb",
         "plot_calibration_curve": "052c_calibration.ipynb",
         "Learner.calibrate_model": "052c_calibration.ipynb",
         "TSClassifier": "052d_tslearner.ipynb",
         "TSRegressor": "052d_tslearner.ipynb",
         "TSForecaster": "052d_tslearner.ipynb",
         "wrap_optimizer": "053_optimizer.ipynb",
         "TransformScheduler": "060_callback.core.ipynb",
         "ShowGraph": "060_callback.core.ipynb",
         "ShowGraphCallback2": "060_callback.core.ipynb",
         "SaveModel": "060_callback.core.ipynb",
         "get_lds_kernel_window": "060_callback.core.ipynb",
         "prepare_LDS_weights": "060_callback.core.ipynb",
         "WeightedPerSampleLoss": "060_callback.core.ipynb",
         "BatchSubsampler": "060_callback.core.ipynb",
         "GamblersCallback": "060_callback.experimental.ipynb",
         "gambler_loss": "060_callback.experimental.ipynb",
         "UBDAug": "060_callback.experimental.ipynb",
         "BatchLossFilter": "060_callback.experimental.ipynb",
         "RandomWeightLossWrapper": "060_callback.experimental.ipynb",
         "BatchMasker": "060_callback.experimental.ipynb",
         "SamplerWithReplacement": "060_callback.experimental.ipynb",
         "NoisyStudent": "061_callback.noisy_student.ipynb",
         "create_subsequence_mask": "063_callback.MVP.ipynb",
         "create_variable_mask": "063_callback.MVP.ipynb",
         "create_future_mask": "063_callback.MVP.ipynb",
         "self_mask": "063_callback.MVP.ipynb",
         "create_mask": "063_callback.MVP.ipynb",
         "MVP": "063_callback.MVP.ipynb",
         "TSBERT": "063_callback.MVP.ipynb",
         "PredictionDynamics": "064_callback.PredictionDynamics.ipynb",
         "init_lin_zero": "100_models.layers.ipynb",
         "lin_zero_init": "100_models.layers.ipynb",
         "SwishBeta": "100_models.layers.ipynb",
         "SmeLU": "100_models.layers.ipynb",
         "Chomp1d": "100_models.layers.ipynb",
         "same_padding1d": "100_models.layers.ipynb",
         "Pad1d": "100_models.layers.ipynb",
         "SameConv1d": "100_models.layers.ipynb",
         "same_padding2d": "100_models.layers.ipynb",
         "Pad2d": "100_models.layers.ipynb",
         "Conv2dSame": "100_models.layers.ipynb",
         "Conv2d": "100_models.layers.ipynb",
         "CausalConv1d": "100_models.layers.ipynb",
         "Conv1d": "100_models.layers.ipynb",
         "SeparableConv1d": "100_models.layers.ipynb",
         "AddCoords1d": "100_models.layers.ipynb",
         "ConvBlock": "100_models.layers.ipynb",
         "Conv": "100_models.layers.ipynb",
         "ConvBN": "100_models.layers.ipynb",
         "CoordConv": "100_models.layers.ipynb",
         "SepConv": "100_models.layers.ipynb",
         "ResBlock1dPlus": "100_models.layers.ipynb",
         "SEModule1d": "100_models.layers.ipynb",
         "Norm": "100_models.layers.ipynb",
         "BN1d": "100_models.layers.ipynb",
         "IN1d": "100_models.layers.ipynb",
         "LinLnDrop": "100_models.layers.ipynb",
         "LambdaPlus": "100_models.layers.ipynb",
         "Squeeze": "100_models.layers.ipynb",
         "Unsqueeze": "100_models.layers.ipynb",
         "Add": "100_models.layers.ipynb",
         "Concat": "100_models.layers.ipynb",
         "Permute": "100_models.layers.ipynb",
         "Transpose": "100_models.layers.ipynb",
         "View": "100_models.layers.ipynb",
         "Reshape": "100_models.layers.ipynb",
         "Max": "100_models.layers.ipynb",
         "LastStep": "100_models.layers.ipynb",
         "SoftMax": "100_models.layers.ipynb",
         "Clamp": "100_models.layers.ipynb",
         "Clip": "100_models.layers.ipynb",
         "ReZero": "100_models.layers.ipynb",
         "Noop": "100_models.layers.ipynb",
         "DropPath": "100_models.layers.ipynb",
         "Sharpen": "100_models.layers.ipynb",
         "Sequential": "122_models.TabFusionTransformer.ipynb",
         "TimeDistributed": "100_models.layers.ipynb",
         "Temp_Scale": "100_models.layers.ipynb",
         "Vector_Scale": "100_models.layers.ipynb",
         "Matrix_Scale": "100_models.layers.ipynb",
         "get_calibrator": "100_models.layers.ipynb",
         "LogitAdjustmentLayer": "100_models.layers.ipynb",
         "LogitAdjLayer": "100_models.layers.ipynb",
         "PPV": "100_models.layers.ipynb",
         "PPAuc": "100_models.layers.ipynb",
         "MaxPPVPool1d": "100_models.layers.ipynb",
         "AdaptiveWeightedAvgPool1d": "100_models.layers.ipynb",
         "GAP1d": "100_models.layers.ipynb",
         "GACP1d": "100_models.layers.ipynb",
         "GAWP1d": "100_models.layers.ipynb",
         "GlobalWeightedAveragePool1d": "100_models.layers.ipynb",
         "gwa_pool_head": "100_models.layers.ipynb",
         "GWAP1d": "100_models.layers.ipynb",
         "AttentionalPool1d": "100_models.layers.ipynb",
         "GAttP1d": "100_models.layers.ipynb",
         "attentional_pool_head": "100_models.layers.ipynb",
         "PoolingLayer": "100_models.layers.ipynb",
         "GEGLU": "123_models.TSPerceiver.ipynb",
         "ReGLU": "100_models.layers.ipynb",
         "get_act_fn": "100_models.layers.ipynb",
         "pytorch_acts": "100_models.layers.ipynb",
         "pytorch_act_names": "100_models.layers.ipynb",
         "create_pool_head": "100_models.layers.ipynb",
         "pool_head": "100_models.layers.ipynb",
         "average_pool_head": "100_models.layers.ipynb",
         "concat_pool_head": "100_models.layers.ipynb",
         "max_pool_head": "100_models.layers.ipynb",
         "create_pool_plus_head": "100_models.layers.ipynb",
         "pool_plus_head": "100_models.layers.ipynb",
         "create_conv_head": "100_models.layers.ipynb",
         "conv_head": "100_models.layers.ipynb",
         "create_mlp_head": "100_models.layers.ipynb",
         "mlp_head": "100_models.layers.ipynb",
         "create_fc_head": "100_models.layers.ipynb",
         "fc_head": "100_models.layers.ipynb",
         "create_rnn_head": "100_models.layers.ipynb",
         "rnn_head": "100_models.layers.ipynb",
         "imputation_head": "100_models.layers.ipynb",
         "create_conv_lin_nd_head": "100_models.layers.ipynb",
         "conv_lin_nd_head": "100_models.layers.ipynb",
         "conv_lin_3d_head": "100_models.layers.ipynb",
         "create_conv_lin_3d_head": "100_models.layers.ipynb",
         "create_lin_nd_head": "100_models.layers.ipynb",
         "lin_nd_head": "100_models.layers.ipynb",
         "lin_3d_head": "100_models.layers.ipynb",
         "create_lin_3d_head": "100_models.layers.ipynb",
         "create_conv_3d_head": "100_models.layers.ipynb",
         "conv_3d_head": "100_models.layers.ipynb",
         "universal_pool_head": "100_models.layers.ipynb",
         "heads": "100_models.layers.ipynb",
         "SqueezeExciteBlock": "100_models.layers.ipynb",
         "GaussianNoise": "100_models.layers.ipynb",
         "PositionwiseFeedForward": "100_models.layers.ipynb",
         "TokenLayer": "100_models.layers.ipynb",
         "ScaledDotProductAttention": "123_models.TSPerceiver.ipynb",
         "MultiheadAttention": "100_models.layers.ipynb",
         "MultiConv1d": "100_models.layers.ipynb",
         "LSTMOutput": "100_models.layers.ipynb",
         "emb_sz_rule": "100_models.layers.ipynb",
         "TSEmbedding": "100_models.layers.ipynb",
         "MultiEmbedding": "100_models.layers.ipynb",
         "apply_idxs": "100b_models.utils.ipynb",
         "SeqTokenizer": "100b_models.utils.ipynb",
         "SeqEmbed": "100b_models.utils.ipynb",
         "get_embed_size": "100b_models.utils.ipynb",
         "get_layers": "100b_models.utils.ipynb",
         "is_layer": "100b_models.utils.ipynb",
         "is_linear": "100b_models.utils.ipynb",
         "is_bn": "100b_models.utils.ipynb",
         "is_conv_linear": "100b_models.utils.ipynb",
         "is_affine_layer": "100b_models.utils.ipynb",
         "is_conv": "100b_models.utils.ipynb",
         "has_bias": "100b_models.utils.ipynb",
         "has_weight": "100b_models.utils.ipynb",
         "has_weight_or_bias": "100b_models.utils.ipynb",
         "check_bias": "100b_models.utils.ipynb",
         "check_weight": "100b_models.utils.ipynb",
         "get_nf": "100b_models.utils.ipynb",
         "ts_splitter": "100b_models.utils.ipynb",
         "transfer_weights": "100b_models.utils.ipynb",
         "build_ts_model": "100b_models.utils.ipynb",
         "build_model": "100b_models.utils.ipynb",
         "create_model": "100b_models.utils.ipynb",
         "count_parameters": "100b_models.utils.ipynb",
         "build_tsimage_model": "100b_models.utils.ipynb",
         "build_tabular_model": "100b_models.utils.ipynb",
         "create_tabular_model": "100b_models.utils.ipynb",
         "get_clones": "100b_models.utils.ipynb",
         "split_model": "100b_models.utils.ipynb",
         "output_size_calculator": "100b_models.utils.ipynb",
         "change_model_head": "100b_models.utils.ipynb",
         "naive_forecaster": "100b_models.utils.ipynb",
         "true_forecaster": "100b_models.utils.ipynb",
         "PositionalEncoding": "100c_models.positional_encoders.ipynb",
         "SinCosPosEncoding": "100c_models.positional_encoders.ipynb",
         "Coord2dPosEncoding": "100c_models.positional_encoders.ipynb",
         "Coord1dPosEncoding": "100c_models.positional_encoders.ipynb",
         "get_acts_and_grads": "100d_models.explainability.ipynb",
         "get_attribution_map": "100d_models.explainability.ipynb",
         "ResBlock": "101_models.ResNet.ipynb",
         "ResNet": "101_models.ResNet.ipynb",
         "ResBlockPlus": "101b_models.ResNetPlus.ipynb",
         "ResNetPlus": "101b_models.ResNetPlus.ipynb",
         "InceptionModule": "102_models.InceptionTime.ipynb",
         "InceptionBlock": "102_models.InceptionTime.ipynb",
         "InceptionTime": "102_models.InceptionTime.ipynb",
         "InceptionModulePlus": "102b_models.InceptionTimePlus.ipynb",
         "InceptionBlockPlus": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus": "102b_models.InceptionTimePlus.ipynb",
         "InCoordTime": "102b_models.InceptionTimePlus.ipynb",
         "XCoordTime": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus17x17": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus32x32": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus47x47": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus62x62": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimeXLPlus": "102b_models.InceptionTimePlus.ipynb",
         "MultiInceptionTimePlus": "102b_models.InceptionTimePlus.ipynb",
         "MLP": "103_models.MLP.ipynb",
         "FCN": "103b_models.FCN.ipynb",
         "FCNPlus": "103c_models.FCNPlus.ipynb",
         "gMLP": "103d_models.gMLP.ipynb",
         "ResCNN": "104_models.ResCNN.ipynb",
         "RNN": "105_models.RNN.ipynb",
         "LSTM": "105_models.RNN.ipynb",
         "GRU": "105_models.RNN.ipynb",
         "RNNPlus": "105_models.RNNPlus.ipynb",
         "LSTMPlus": "105_models.RNNPlus.ipynb",
         "GRUPlus": "105_models.RNNPlus.ipynb",
         "XceptionModule": "106_models.XceptionTime.ipynb",
         "XceptionBlock": "106_models.XceptionTime.ipynb",
         "XceptionTime": "106_models.XceptionTime.ipynb",
         "XceptionModulePlus": "106b_models.XceptionTimePlus.ipynb",
         "XceptionBlockPlus": "106b_models.XceptionTimePlus.ipynb",
         "XceptionTimePlus": "106b_models.XceptionTimePlus.ipynb",
         "RNN_FCN": "107_models.RNN_FCN.ipynb",
         "LSTM_FCN": "107_models.RNN_FCN.ipynb",
         "GRU_FCN": "107_models.RNN_FCN.ipynb",
         "MRNN_FCN": "107_models.RNN_FCN.ipynb",
         "MLSTM_FCN": "107_models.RNN_FCN.ipynb",
         "MGRU_FCN": "107_models.RNN_FCN.ipynb",
         "RNN_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "LSTM_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "GRU_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "MRNN_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "MLSTM_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "MGRU_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "TransformerModel": "108_models.TransformerModel.ipynb",
         "get_activation_fn": "108b_models.TST.ipynb",
         "TST": "108b_models.TST.ipynb",
         "TSTPlus": "108c_models.TSTPlus.ipynb",
         "MultiTSTPlus": "108c_models.TSTPlus.ipynb",
         "SampaddingConv1D_BN": "109_models.OmniScaleCNN.ipynb",
         "build_layer_with_layer_parameter": "109_models.OmniScaleCNN.ipynb",
         "OmniScaleCNN": "109_models.OmniScaleCNN.ipynb",
         "get_Prime_number_in_a_range": "109_models.OmniScaleCNN.ipynb",
         "get_out_channel_number": "109_models.OmniScaleCNN.ipynb",
         "generate_layer_parameter_list": "109_models.OmniScaleCNN.ipynb",
         "WaveBlock": "110_models.mWDN.ipynb",
         "mWDN": "110_models.mWDN.ipynb",
         "mWDNBlocks": "110_models.mWDN.ipynb",
         "mWDNPlus": "110_models.mWDN.ipynb",
         "RocketClassifier": "111_models.ROCKET.ipynb",
         "load_rocket": "111_models.ROCKET.ipynb",
         "RocketRegressor": "111_models.ROCKET.ipynb",
         "ROCKET": "111a_models.ROCKET_Pytorch.ipynb",
         "create_rocket_features": "111a_models.ROCKET_Pytorch.ipynb",
         "get_rocket_features": "111a_models.ROCKET_Pytorch.ipynb",
         "MiniRocketClassifier": "111b_models.MINIROCKET.ipynb",
         "load_minirocket": "111b_models.MINIROCKET.ipynb",
         "MiniRocketRegressor": "111b_models.MINIROCKET.ipynb",
         "MiniRocketVotingClassifier": "111b_models.MINIROCKET.ipynb",
         "get_minirocket_preds": "111b_models.MINIROCKET.ipynb",
         "MiniRocketVotingRegressor": "111b_models.MINIROCKET.ipynb",
         "MiniRocketFeatures": "111c_models.MINIROCKET_Pytorch.ipynb",
         "MRF": "111c_models.MINIROCKET_Pytorch.ipynb",
         "get_minirocket_features": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "MiniRocketHead": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "MiniRocket": "111c_models.MINIROCKET_Pytorch.ipynb",
         "MiniRocketFeaturesPlus": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "Flatten": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "MiniRocketPlus": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "InceptionRocketFeaturesPlus": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "InceptionRocketPlus": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "xresnet1d18": "112_models.XResNet1d.ipynb",
         "xresnet1d34": "112_models.XResNet1d.ipynb",
         "xresnet1d50": "112_models.XResNet1d.ipynb",
         "xresnet1d101": "112_models.XResNet1d.ipynb",
         "xresnet1d152": "112_models.XResNet1d.ipynb",
         "xresnet1d18_deep": "112_models.XResNet1d.ipynb",
         "xresnet1d34_deep": "112_models.XResNet1d.ipynb",
         "xresnet1d50_deep": "112_models.XResNet1d.ipynb",
         "xresnet1d18_deeper": "112_models.XResNet1d.ipynb",
         "xresnet1d34_deeper": "112_models.XResNet1d.ipynb",
         "xresnet1d50_deeper": "112_models.XResNet1d.ipynb",
         "XResNet1dPlus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d18plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d34plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d50plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d101plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d152plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d18_deepplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d34_deepplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d50_deepplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d18_deeperplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d34_deeperplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d50_deeperplus": "112b_models.XResNet1dPlus.ipynb",
         "TemporalBlock": "113_models.TCN.ipynb",
         "TemporalConvNet": "113_models.TCN.ipynb",
         "TCN": "113_models.TCN.ipynb",
         "XCM": "114_models.XCM.ipynb",
         "XCMPlus": "114b_models.XCMPlus.ipynb",
         "TabModel": "120_models.TabModel.ipynb",
         "TabBackbone": "120_models.TabModel.ipynb",
         "TabHead": "120_models.TabModel.ipynb",
         "ifnone": "122_models.TabFusionTransformer.ipynb",
         "SharedEmbedding": "121_models.TabTransformer.ipynb",
         "FullEmbeddingDropout": "121_models.TabTransformer.ipynb",
         "TabTransformer": "121_models.TabTransformer.ipynb",
         "GatedTabTransformer": "121b_models.GatedTabTransformer.ipynb",
         "TabFusionBackbone": "122_models.TabFusionTransformer.ipynb",
         "TabFusionTransformer": "122_models.TabFusionTransformer.ipynb",
         "TSTabFusionTransformer": "122_models.TabFusionTransformer.ipynb",
         "Attention": "123_models.TSPerceiver.ipynb",
         "FeedForward": "123_models.TSPerceiver.ipynb",
         "CrossAttention": "123_models.TSPerceiver.ipynb",
         "LatentTransformer": "123_models.TSPerceiver.ipynb",
         "TSPerceiver": "123_models.TSPerceiver.ipynb",
         "TSiTPlus": "124_models.TSiTPlus.ipynb",
         "TSiT": "124_models.TSiTPlus.ipynb",
         "TSSequencerPlus": "125_models.TSSequencerPlus.ipynb",
         "TSSequencer": "125_models.TSSequencerPlus.ipynb",
         "MultiInputNet": "130_models.MultiInputNet.ipynb",
         "InputWrapper": "140_models.misc.ipynb",
         "ResidualWrapper": "140_models.misc.ipynb",
         "RecursiveWrapper": "140_models.misc.ipynb",
         "run_optuna_study": "200_optuna.ipynb",
         "wandb_agent": "201_wandb.ipynb",
         "get_wandb_agent": "201_wandb.ipynb",
         "run_wandb_agent": "201_wandb.ipynb"}

modules = ["export.py",
           "utils.py",
           "data/validation.py",
           "data/preparation.py",
           "data/external.py",
           "data/core.py",
           "data/unwindowed.py",
           "data/metadatasets.py",
           "data/preprocessing.py",
           "data/transforms.py",
           "data/mixed_augmentation.py",
           "data/image.py",
           "data/features.py",
           "data/tabular.py",
           "data/mixed.py",
           "losses.py",
           "metrics.py",
           "learner.py",
           "inference.py",
           "analysis.py",
           "calibration.py",
           "tslearner.py",
           "optimizer.py",
           "callback/core.py",
           "callback/experimental.py",
           "callback/noisy_student.py",
           "callback/MVP.py",
           "callback/PredictionDynamics.py",
           "models/layers.py",
           "models/utils.py",
           "models/positional_encoders.py",
           "models/explainability.py",
           "models/ResNet.py",
           "models/ResNetPlus.py",
           "models/InceptionTime.py",
           "models/InceptionTimePlus.py",
           "models/MLP.py",
           "models/FCN.py",
           "models/FCNPlus.py",
           "models/gMLP.py",
           "models/ResCNN.py",
           "models/RNN.py",
           "models/RNNPlus.py",
           "models/XceptionTime.py",
           "models/XceptionTimePlus.py",
           "models/RNN_FCN.py",
           "models/RNN_FCNPlus.py",
           "models/TransformerModel.py",
           "models/TST.py",
           "models/TSTPlus.py",
           "models/OmniScaleCNN.py",
           "models/mWDN.py",
           "models/ROCKET.py",
           "models/ROCKET_Pytorch.py",
           "models/MINIROCKET.py",
           "models/MINIROCKET_Pytorch.py",
           "models/MINIROCKETPlus_Pytorch.py",
           "models/XResNet1d.py",
           "models/XResNet1dPlus.py",
           "models/TCN.py",
           "models/XCM.py",
           "models/XCMPlus.py",
           "models/TabModel.py",
           "models/TabTransformer.py",
           "models/GatedTabTransformer.py",
           "models/TabFusionTransformer.py",
           "models/TSPerceiver.py",
           "models/TSiTPlus.py",
           "models/TSSequencerPlus.py",
           "models/MultiInputNet.py",
           "models/misc.py",
           "optuna.py",
           "wandb.py"]

doc_url = "https://timeseriesAI.github.io/tsai/"

git_url = "https://github.com/timeseriesAI/tsai/tree/main/"

def custom_doc_links(name): return None
